(*   globals.ml
*)

type expr
    = NUM  of int    (*number*)
    | BOOL of bool   (*Boolean*)
    | VAR  of string (*variable*)
    (*arithmetic exprs*)
    | ADD of expr * expr | SUB of expr * expr | MUL of expr * expr | DIV of expr * expr
    (*comparators*)
    | EQ of expr * expr  | GE of expr * expr
    (*logical exprs*)
    | AND of expr * expr | OR  of expr * expr | NOT of expr
    (*conditional expr*)
    | IF  of expr * expr * expr
    (*function definition: parameter, body*)
    | FUN of string * expr
    (*closure: parameter, body, environment*)
    | CLO of string * expr * (string * expr) list
    (*lazy eval, thunk: expr, env*)
    | TNK of expr * (string * expr) list
    (*function application: operator, operand*)
    | APP of expr * expr

module type IStateMonad = sig
    type 'a t
    type state

    (*monad*)
    val ret : 'a -> 'a t
    val (>>=) : 'a t -> ('a -> 'b t) -> 'b t

    (*state monad*)
    val get : state t
    val set : state -> unit t
end

module StateMonad (State: sig type t end)
    : (IStateMonad with type state = State.t and type 'a t = State.t -> ('a * State.t))
= struct

    (*monad operations*)
    type state = State.t
    type 'a t = state -> ('a * state)

    (*return: wraps values in a monad type*)
    let ret value =
        fun state -> (value, state)

    (*bind: composes functions that return a monad type*)
    let (>>=) mon f =
        fun state ->
            mon state |> fun (value, state') -> (f value) state'
    
    (*other operations for state monad*)
    let get =
        fun state -> (state, state)

    let set state =
        fun state' -> ((), state)
end
